/*
flip51   Software for the ELV Flash Programmer FP51
Copyright (C) 1999-2003 Oliver Tscherwitschke <oliver@tscherwitschke.de>
        
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                        
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
          
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/                                                    
                                                    

/* check if target os is defined */
#ifndef TARGET_WINDOWS
#ifndef TARGET_LINUX
    #error Please define either TARGET_LINUX or TARGET_WINDOWS
#endif
#endif

/* include files */
#include <stdio.h>
#ifdef TARGET_LINUX
    #include <unistd.h>
    #include <asm/io.h>
#endif
#ifdef TARGET_WINDOWS
    #include <windows.h>
    #include <conio.h>
#endif

#include <string.h>

#include "types.h"
#include "flip51.h"
#include "io.h"
#include "ihex.h"
#include "delay.h"
#include "tools.h"



/* global variables */
unsigned short  baseport = 0x378;           /* i/o adress of parallel port */

int     do_erase = 0;       /* what to do? */
int     do_program = 0;
int     do_program_hex = 0;
int     do_read = 0;
int     do_verify = 0;
int     do_verify_hex = 0;
int     do_lockbit = 0;

int     lockbit[3] = {0, 0, 0};

char    *infilename;
char    *outfilename;
char    *vfilename;

#ifdef TARGET_WINDOWS
int     do_testspeed = 0;
int     use_giveio = 0;
int		install_giveio = 0;
int		remove_giveio = 0;
int     giveio_installed = 0;
int     giveio_started = 0;
char	giveio_path[256];
char	giveio_name[] = "GIVEIO";
char*	giveio_filename;
DWORD   error;
#endif


/****************************************************************/
/* reset all internal buffers of the FP51                       */
/* and set parallel port lines to default values                */
/*                                                              */
/****************************************************************/
void fp51_reset()
{ 
    OUTB(8 + 4, baseport + 2);                      /* init control reg */
    USLEEP(MS1);
    
    /* reset all latches */
    OUTB(INB(baseport + 2) & ~4, baseport + 2);     /* set init low     */
    USLEEP(MS1);
    OUTB(INB(baseport + 2) | 4, baseport + 2);      /* set init to high */
    USLEEP(MS1);
    
    write_byte_to_latch(0x81, 3);                    /* bit 0 and 7 high, RST low */
    USLEEP(MS1);
    OUTPUT_DISABLE;
}


/****************************************************************/
/* check if FP51-hardware is present.                           */
/* returns 1, if fp51 was found, otherwise 0                    */
/****************************************************************/
int fp51_check()
{
    OUTB(INB(baseport + 2) & ~8, baseport + 2);  /* set select lo */
    USLEEP(IOWAIT);
    if (!(INB(baseport + 1) & 64))
        return 0;

    OUTB(INB(baseport + 2) | 8, baseport + 2);      /* set select high */
    USLEEP(IOWAIT);
    if ((INB(baseport + 1) & 64))
        return 0;

    return 1;
}


/****************************************************************/
/* erase the flash                                              */
/****************************************************************/
void chip_erase(CHIP_INFO *chipinfo)
{
    /* power on sequence */
    VCC_OFF;
    VPP_12V_OFF;
    VPP_5V_OFF;
    PSEN_LOW;                           /* clear clock line */
    RST_LOW;
    set_mode(0);                        /* idle mode */
    USLEEP(MS100);
    VCC_ON;                             
    RST_HIGH;
    VPP_5V_ON;
    PROG_HIGH;
    USLEEP(MS100);
    

    set_mode(0x08);                 /* chipo erase */
    USLEEP(IOWAIT);

    if (chipinfo->volt == 12)
        VPP_12V_ON;
    USLEEP(IOWAIT);

    PROG_LOW;
    USLEEP(MS10);                   /* generate 10ms prog pulse */
    PROG_HIGH;
    USLEEP(IOWAIT);

    if (chipinfo->volt == 12)
        VPP_12V_OFF;
    USLEEP(IOWAIT);
    
    
    /* power off */
    VPP_12V_OFF;
    VPP_5V_OFF;
    PSEN_LOW;                           /* clear clock line */
    PROG_LOW;
    VCC_OFF;
    RST_LOW;
}


/****************************************************************/
/* read the signature bytes of the uC                           */
/* return value:                                                */
/*         0:   error                                           */
/*      0x11:   1051                                            */
/*      0x21:   2051                                            */
/*      0x41:   4051                                            */
/*      0x51:   51                                              */
/*      0x52:   52                                              */
/*                                                              */
/****************************************************************/
int read_sig(CHIP_INFO *chipinfo)
{
    byte s1, s2, s3;

    VCC_ON;

    /* generate one complete reset pulse */
    RST_HIGH;                           /* for 40p */
    VPP_5V_ON;                          /* = RST for 20p */
    USLEEP(MS1);
    RST_LOW;                           /* for 40p */
    VPP_5V_OFF;                          /* = RST for 20p */
    USLEEP(MS1);
    RST_HIGH;                           /* for 40p */
    VPP_5V_ON;                          /* = RST for 20p */

    PROG_HIGH;
    PSEN_LOW;                           /* only needed for 40p */
    set_mode(4);                        /* read sig mode */
    USLEEP(IOWAIT);

    /* now read the 3 signature bytes. They have the addresses 0, 1, 2
       for the 20 pin chips and 0x30, 0x31 and 0x32 for the 40 pin chips.
       We address only 0x30 ... 0x32 because the 20pin chips dont have
       address lines.                           
    */

    write_byte_to_latch(latch13 & 0xe0, 2);         /* high-byte of adr = 0 */      
    write_byte_to_latch(0x30, 1);                   /* address 0x30 */ 
    set_mode(0);                                    /* enable low */
    USLEEP(DRWAIT);                                     /* wait 12 us for data ready */
    s1 = read_byte();                               /* read first sig. byte */
    set_mode(4);
    USLEEP(DRWAIT);                                     /* wait 12 us for data ready */
    PSEN_HIGH;                                      /* 20p only: */
    USLEEP(1);
    PSEN_LOW;                                       /* inc address counter */
    
    write_byte_to_latch(0x31, 1);                   /* address 0x31 */
    set_mode(0);                                    /* enable low */
    USLEEP(DRWAIT);
    s2 = read_byte();                               /* read 2. sig. byte */
    set_mode(4);
    USLEEP(DRWAIT);                                     /* wait 12 us for data ready */
    PSEN_HIGH;                                      /* 20p only: */
    USLEEP(1);
    PSEN_LOW;                                       /* inc address counter */

    write_byte_to_latch(0x32, 1);                   /* address 0x32 */
    set_mode(0);                                    /* enable low */
    USLEEP(DRWAIT);
    s3 = read_byte();                               /* read 3. sig. byte */
    set_mode(0);

    /* switch everything off */
    write_byte_to_latch(0x0, 1);                    /* address 0 */
    PROG_LOW;
    VPP_5V_OFF;
    VCC_OFF;    
    RST_LOW;

    if ((s1 != 0x1E) && (revb(s1) != 0x1E))
        return 0;                                   /* no atmel chip */

    if (revb(s2) == 0x11)
    {
        chipinfo->type = 0x11;                      /* 1051 */
        chipinfo->size = 1024;
        strcpy(chipinfo->chipname, "AT89C1051");
    }
    else if (revb(s2) == 0x21)
    {
        chipinfo->type = 0x21;                      /* 2051 */
        chipinfo->size = 2048;
        strcpy(chipinfo->chipname, "AT89C2051");
    }
    else if (revb(s2) == 0x41)
    {
        chipinfo->type = 0x41;                       /* 4051 */
        chipinfo->size = 4096;
        strcpy(chipinfo->chipname, "AT89C4051");
    }
    else if (s2 == 0x51)
    {
        chipinfo->type = 0x51;                      /* 51 */
        chipinfo->size = 4096;
        strcpy(chipinfo->chipname, "AT89C51");
    }
    else if (s2 == 0x52)
    {
        chipinfo->type = 0x52;                      /* 52 */
        chipinfo->size = 8192;
        strcpy(chipinfo->chipname, "AT89C52");
    }

    if (s3 == 0xFF)
        chipinfo->volt = 12;                        /* Vpp = 12V */
    else if ((chipinfo->type >= 0x51 && s3 == 0x05) || (chipinfo->type < 0x51 && revb(s3) == 0x05))
        chipinfo->volt = 5;                         /* Vpp = 5V */

    if (chipinfo->type >= 0x51)
    {
        chipinfo->sig[0] = s1;
        chipinfo->sig[1] = s2;
        chipinfo->sig[2] = s3;
    }
    else
    {
        chipinfo->sig[0] = revb(s1);
        chipinfo->sig[1] = revb(s2);
        chipinfo->sig[2] = revb(s3);        
    }    

    return chipinfo->type;
}


/****************************************************************/
/* read one byte from the uC                                    */
/*  ATTENTION: Data from a 20 pin uC will be                    */
/*  bit-reversed, i.e. 01234567 instead of                      */
/*  76543210. Use revb() to reverse a byte                      */
/****************************************************************/
byte read_byte()
{
    byte b1, b2;

    latch_7((latch7 & 0xf8) | 5);
    OUTB(INB(baseport + 2) | 2, baseport + 2);      /* enable address-decoder */                    
    USLEEP(DRWAIT);
    b1 = INB(baseport + 1);                         /* high nibble */
    OUTB(INB(baseport + 2) & ~2, baseport + 2);     /* disable address-decoder */   

    b1 = b1 >> 3;                                   /* shift to the right */
    b1 &= 0x0f;                                     /* isolate bits */
    b1 ^= 0x0f;                                     /* invert all bits */

    latch_7((latch7 & 0xf8) | 4);
    OUTB(INB(baseport + 2) | 2, baseport + 2);      /* enable address-decoder */                    
    USLEEP(DRWAIT);
    b2 = INB(baseport + 1);                         /* low nibble */
    OUTB(INB(baseport + 2) & ~2, baseport + 2);     /* disable address-decoder */   
    b2 = b2 << 1;                                   /* shift to the left */
    b2 &= 0xf0;                                     /* isolate bits */
    b2 ^= 0xf0;                                     /* invert all bits */

    return b1 | b2;
}


/****************************************************************/
/* read the entire flash of the uC an store it in               */
/* the specified buffer                                         */
/****************************************************************/
void read_flash(byte *buffer, CHIP_INFO *chipinfo, int readsize)
{
    int i;
    byte s;
    
    if (readsize > chipinfo->size)
        return;
    
    VCC_ON;

    /* generate one complete reset pulse */
    RST_HIGH;                           /* for 40p */
    VPP_5V_ON;                          /* = RST for 20p */
    USLEEP(MS1);
    RST_LOW;                           /* for 40p */
    VPP_5V_OFF;                          /* = RST for 20p */
    USLEEP(MS1);
    RST_HIGH;                           /* for 40p */
    VPP_5V_ON;                          /* = RST for 20p */
    
    PROG_HIGH;
    PSEN_LOW;   
    set_mode(3 + 4);                        /* read mode */
    USLEEP(IOWAIT);

    for (i = 0; i < readsize; i++) {
        if (chipinfo->type == 0x51 || chipinfo->type == 0x52) {     /* address a byte (40pin only) */
            write_byte_to_latch((latch13 & 0xe0) | (i >> 8), 2);    /* high-byte of adr */      
            write_byte_to_latch(i & 0xff, 1);           /* low-byte */
        }
        USLEEP(DRWAIT);                     /* wait for data ready */
        set_mode(3);
        USLEEP(DRWAIT);                     /* wait for data ready */
        s = read_byte();                /* read first sig. byte */
        set_mode(3 + 4);
        USLEEP(DRWAIT);                     /* wait for data ready */

        if (chipinfo->type == 0x11 || chipinfo->type == 0x21 || chipinfo->type == 0x41) {
            s = revb(s);
            PSEN_HIGH;              /* inc addr. counter */
            USLEEP(1);
            PSEN_LOW;
        }
        buffer[i] = s;
    }
    /* switch everything off */
    write_byte_to_latch(0x0, 1);                /* address 0 */
    PROG_LOW;
    VPP_5V_OFF;
    RST_LOW;
    VCC_OFF;
    set_mode(0);    
    write_byte_to_latch((latch13 & 0xe0), 2);       /* high-byte of adr = 0 */      
    write_byte_to_latch(0, 1);
}


/****************************************************************/
/* Write 'size' bytes from 'buffer' in flash rom                */
/****************************************************************/
int write_flash(byte *buffer, CHIP_INFO *chipinfo, int size)
{
    int i, error = 0, percent, oldpercent = -1;
    byte data;
        
    VCC_ON;
    PSEN_LOW;
    USLEEP(MS1);
    if (chipinfo->type == 0x11 || chipinfo->type == 0x21 || chipinfo->type == 0x41) {         /* power up sequence */
        VPP_12V_OFF;
        VPP_5V_OFF;
        USLEEP(20000);
        VPP_5V_ON;
    }
    RST_HIGH;                       /* for 40p */
    PROG_HIGH;
    USLEEP(MS1);
    
    for (i = 0; i < size; i++) {
        OUTPUT_ENABLE;
        set_mode(7);
        USLEEP(IOWAIT);
        
        /* address a byte (40pin only) */
        if (chipinfo->type == 0x51 || chipinfo->type == 0x52) {
            write_byte_to_latch((latch13 & 0xe0) | (i >> 8), 2);    /* high-byte of adr */      
            write_byte_to_latch(i & 0xff, 1);                       /* low-byte */
            USLEEP(DRWAIT);
            
            /* byte to write */
            write_byte_to_latch(buffer[i], 6);
            USLEEP(DRWAIT);
        } else
            write_byte_to_latch(revb(buffer[i]), 6);
            USLEEP(DRWAIT);
            
        if (chipinfo->volt == 12){
            VPP_12V_ON;
            VPP_5V_OFF;
        } else if (chipinfo->volt == 5) {
            VPP_5V_ON;
            VPP_12V_OFF;
         }
        USLEEP(DRWAIT);

        /* pulse prog to program one byte */
        write_byte_to_latch(latch14 & 0xfe, 3);
        USLEEP(200);
        write_byte_to_latch(latch14 | 0x1, 3);
        USLEEP(IOWAIT);
        wait_for_ready();               /* wait for ready */
        percent = (i*100)/size;
        if (percent != oldpercent) { 
            oldpercent = percent;
            printf("%i%\x25\r", percent);
            fflush(stdout);
        }
        
        OUTPUT_DISABLE;                 /* disable driver ic9 */
                                         /* to be able to read a byte */
        if (chipinfo->volt == 12) {
            VPP_5V_ON;
            VPP_12V_OFF;
        }
        set_mode(3);                    /* verify */
        USLEEP(DRWAIT);
        data = read_byte();             /* read byte */
        if (chipinfo->type == 0x11 || chipinfo->type == 0x21 || chipinfo->type == 0x41) {
            data = revb(data);
            PSEN_HIGH;                  /* inc addr. counter */                 /* 20p only: */
            PSEN_LOW;
        }
        if (data != buffer[i]) {
            printf("error: writing byte %i not successful: wrote %X, read %X\n", i, buffer[i], data);
            error = 1;
            break;                      /* exit programming loop */
        }
    }
    
    /* switch everything off */
    write_byte_to_latch(0x0, 1);                /* address 0 */
    PROG_LOW;
    VPP_5V_OFF;
    VPP_12V_OFF;
    RST_LOW;
    VCC_OFF;
    set_mode(0);    
    write_byte_to_latch((latch13 & 0xe0), 2);       /* high-byte of adr = 0 */      
    write_byte_to_latch(0, 1);

    return error;
}


/****************************************************************/
/* program the lockbits                                         */
/****************************************************************/
void prog_lockbits(CHIP_INFO *chipinfo)
{
    /* power on sequence */
    VCC_OFF;
    VPP_12V_OFF;
    VPP_5V_OFF;
    PSEN_LOW;                           /* clear clock line */
    RST_LOW;
    set_mode(0);                        /* idle mode */
    USLEEP(MS100);
    VCC_ON;                             
    RST_HIGH;
    VPP_5V_ON;
    PROG_HIGH;
    USLEEP(MS100);
    

    if (lockbit[0]) {
        set_mode(0x0f);                 /* lockbit 1 */
        USLEEP(IOWAIT);

        if (chipinfo->volt == 12)
            VPP_12V_ON;
        USLEEP(IOWAIT);

        PROG_LOW;
        USLEEP(IOWAIT);
        PROG_HIGH;
        USLEEP(IOWAIT);

        if (chipinfo->volt == 12)
            VPP_12V_OFF;
        USLEEP(IOWAIT);
    }
    
    if (lockbit[1]) {
        set_mode(0x0c);                 /* lockbit 1 */
        USLEEP(IOWAIT);

        if (chipinfo->volt == 12)
            VPP_12V_ON;
        USLEEP(IOWAIT);

        PROG_LOW;
        USLEEP(IOWAIT);
        PROG_HIGH;
        USLEEP(IOWAIT);

        if (chipinfo->volt == 12)
            VPP_12V_OFF;
        USLEEP(IOWAIT);
    }   

    if (lockbit[2]) {
        if (chipinfo->type == 0x11 || chipinfo->type == 0x21 || chipinfo->type == 0x41)
            printf("error: uC doesn't have lockbit 3\n");
        else {
            set_mode(0x0a);                 /* lockbit 1 */
            USLEEP(IOWAIT);
    
            if (chipinfo->volt == 12)
                VPP_12V_ON;
            USLEEP(IOWAIT);
    
            PROG_LOW;
            USLEEP(IOWAIT);
            PROG_HIGH;
            USLEEP(IOWAIT);
    
            if (chipinfo->volt == 12)
                VPP_12V_OFF;
            USLEEP(IOWAIT);
        }
    }
    
    
    /* power off */
    set_mode(0);                        
    RST_LOW;
    VPP_5V_OFF;
    PROG_LOW;
    VCC_OFF;
}


/****************************************************************/
/* check the arguments of main()                                */
/* if no args were give, return 1, else return 0                */
/****************************************************************/
int arguments(int argc, char* argv[])
{
    int i;
    char* ioerr;
    int result = 0; 

    if (argc == 1)
        return 1;
        
    for (i = 1; i < argc; i++) {
        if (!strcmp(argv[i], "-e"))
            do_erase = 1;
        else if (!strcmp(argv[i], "-io")) {
            if (argc == (i+1)) {
                printf("error: -io needs argument.\n");
                result = 1;
            }
            baseport = (unsigned short) strtol(argv[i+1], &ioerr, 16);       
            if (*ioerr != 0) {
                printf("error: -io needs numeric argument.\n");
                result = 1;
            }
        } else if (!strcmp(argv[i], "-p")) {
            if (argc > (i+1)) {
                do_program = 1;
                do_program_hex = 0;
                infilename = argv[i+1];
            } else {
                printf("error: -p needs filename.\n");
                result = 1;
            }
        } else if (!strcmp(argv[i], "-ph")) {
            if (argc > (i+1)) {
                do_program = 0;
                do_program_hex = 1;
                infilename = argv[i+1];
            } else {
                printf("error: -ph needs filename.\n");
                result = 1;
            }
        } else if (!strcmp(argv[i], "-r")) {
            if (argc > (i+1)) {
                do_read = 1;
                outfilename = argv[i+1];
            } else {
                printf("error: -r needs filename.\n");
                result = 1;
            }
        } else if (!strcmp(argv[i], "-v")) {
            if (argc > (i+1)) {
                do_verify = 1;
                do_verify_hex = 0;
                vfilename = argv[i+1];
            } else {
                printf("error: -v needs filename.\n");
                result = 1;
            }
        } else if (!strcmp(argv[i], "-vh")) {
            if (argc > (i+1)) {
                do_verify_hex = 1;
                do_verify = 0;
                vfilename = argv[i+1];
            } else {
                printf("error: -v needs filename.\n");
                result = 1;
            }
        } else if (!strcmp(argv[i], "-l")) {
            if (argc > (i+1)) {
                do_lockbit = 1;                                                         
                if (strchr(argv[i+1], '1')) 
                    lockbit[0] = 1;
                if (strchr(argv[i+1], '2'))
                    lockbit[1] = 1;
                if (strchr(argv[i+1], '3'))
                    lockbit[2] = 1;
            } else {
                printf("error: -l needs parameter.\n");
                result = 1;
            }
#ifdef TARGET_WINDOWS
        } else if (!strcmp(argv[i], "-install")) {
            install_giveio = 1;
        } else if (!strcmp(argv[i], "-remove")) {
            remove_giveio = 1;
        } else if (!strcmp(argv[i], "-t")) {
            do_testspeed = 1;
#endif
        } else if (!strcmp(argv[i], "--help")) {
            result = 1;
        }   
    }
    return result;
}


/****************************************************************/
/* read the binary file into the buffer                         */
/* return: size of file or -1 if size to big                    */
/****************************************************************/
long read_bin_file(byte *buffer, unsigned long flashsize, FILE *infile)
{
    unsigned long filesize;
    
    /* read file into buffer */
    fseek(infile, 0, SEEK_END);
    filesize = ftell(infile);
    fseek(infile, 0, SEEK_SET);         
    
    if (filesize > flashsize)           /* file too big */
        return -1;
    else
        fread(buffer, 1, filesize, infile);
    
    return filesize;
}


/****************************************************************/
/* read the intel hex file into the buffer                      */
/* return: size of file or -1 if size to big or other error     */
/****************************************************************/
long read_hex_file(byte *buffer, unsigned long flashsize, FILE *infile)
{
	byte    sBuff[1024];
	byte    rc;
	unsigned long usiMaxAddr = 0;
	unsigned long usiActAddr;
    byte    *c;
	
	do
	{
		/* read one record from infile */
		if (fgets(sBuff, sizeof sBuff, infile) == NULL)
		    break;

        /* remove newline */
        c = sBuff;
        while ((*c != '\n') && (*c != '\r') && (*c != '\0'))
            c++;
        *c = '\0';            
        
		
		if (ihex_checksum(sBuff) == 1)	    /* checksum error */
		{
			printf("Checksum error\n");
			return -1;
		}
		else								/* checksum is OK */
		{
			rc = ihex_load_record(sBuff, buffer, flashsize, &usiActAddr);	/* load record to buffer */
            if (rc == 0)
            {
			    if (usiActAddr > usiMaxAddr)				    /* remember max address */
				    usiMaxAddr = usiActAddr;
            }				
			else if (rc == 2)                                   /* ERR_UNSUPPORTED */
			{
				printf("Unsupported Hex record\n");
				return -1;
			} 
			else if (rc == 3)
			{
				printf("Address out of range\n");
				return -1;
			}
			
		}

	} 
	while (rc != 1);

	return usiMaxAddr + 1;
}


/****************************************************************/
/* Install GIVEIO.SYS service                                   */
/* return: 0   : Success                                        */
/*         else: Error. In addition the global variable 'error' */
/*               is set with a specific error code.             */
/****************************************************************/
int PortDriverInstall(u_char *path)
{
     SC_HANDLE manager;
     SC_HANDLE service;
     u_char *driver = giveio_name;

     if((manager = OpenSCManager(NULL, NULL,  SC_MANAGER_ALL_ACCESS)) == NULL) {
         error = GetLastError();
         return -1;
     }
     if((service = CreateService(manager, driver, driver, SERVICE_ALL_ACCESS,
         SERVICE_KERNEL_DRIVER, SERVICE_AUTO_START, SERVICE_ERROR_NORMAL, path,
         NULL, NULL, NULL, NULL, NULL)) == NULL) {
         error = GetLastError();
         CloseServiceHandle(manager);
         return -2;
     }
     CloseServiceHandle(service);
     CloseServiceHandle(manager);

     return 0;
}


/****************************************************************/
/* Remove GIVEIO.SYS service                                    */
/* return: 0   : Success                                        */
/*         else: Error. In addition the global variable 'error' */
/*               is set with a specific error code.             */
/****************************************************************/
int PortDriverRemove(void)
{
     SC_HANDLE manager;
     SC_HANDLE service;

     if((manager = OpenSCManager(NULL, NULL,  SC_MANAGER_ALL_ACCESS)) == NULL) {
         error = GetLastError();
         return -1;
     }
     if((service = OpenService(manager, giveio_name, SERVICE_ALL_ACCESS)) == NULL) {
         error = GetLastError();
         CloseServiceHandle(manager);
         return -2;
     }
     if(!DeleteService(service)) {
         error = GetLastError();
         return -3;
     }
     CloseServiceHandle(service);
     CloseServiceHandle(manager);

     return 0;
}

/****************************************************************/
/* Start GIVEIO.SYS service                                     */
/* return: 0   : Success                                        */
/*         else: Error. In addition the global variable 'error' */
/*               is set with a specific error code.             */
/****************************************************************/
int PortDriverStart(void)
{
	SC_HANDLE manager;
	SC_HANDLE service;
	BOOL brc;


    if((manager = OpenSCManager(NULL, NULL,  SC_MANAGER_ALL_ACCESS)) == NULL) {
		error = GetLastError();
        return -1;
    }
	
	service = OpenService(manager, giveio_name, SERVICE_START);
	if (service == NULL) {
         error = GetLastError();
         return -2;
    }

	brc = StartService(service, 0, NULL);	
	if (brc == 0) {
         error = GetLastError();
         return -3;
    }

	return 0;
}



/****************************************************************/
/* Stop GIVEIO.SYS service                                      */
/* return: 0   : Success                                        */
/*         else: Error. In addition the global variable 'error' */
/*               is set with a specific error code.             */
/****************************************************************/
int PortDriverStop(void)
{
	SC_HANDLE manager;
	SC_HANDLE service;
	BOOL brc;
	SERVICE_STATUS service_stat;


    if((manager = OpenSCManager(NULL, NULL,  SC_MANAGER_ALL_ACCESS)) == NULL) {
		error = GetLastError();
        return -1;
    }

	service = OpenService(manager, giveio_name, SERVICE_STOP);
	if (service == NULL) {
         error = GetLastError();
         return -2;
    }

	brc = ControlService(service, SERVICE_CONTROL_STOP, &service_stat);
	if (brc == 0) {
         error = GetLastError();
         return -3;
    }

	return 0;
}


/****************************************************************/
/* Oh! A main() function. Cool.                                 */
/****************************************************************/
int main(int argc, char* argv[])
{
#ifdef TARGET_LINUX
    __uid_t olduid;                         /* old user id */
#endif
    byte buffer[8192];
    byte vbuffer[8192];                     /* buffers */
    int rc;    
#ifdef TARGET_WINDOWS
    HANDLE giveio_handle;
	OSVERSIONINFO windows_version;
#endif

    FILE *infile, *outfile;                 /* in- and outbut files */
    unsigned long filesize;
    char chipname[20] = "unknown";          /* name of the uC */
    CHIP_INFO   chipinfo = {0};             /* chip info */


    if (arguments(argc, argv)) {
        printf("flip51 " FLIPVERSION " (c) 1999-2003 Oliver Tscherwitschke <oliver@tscherwitschke.de>\n");
        printf("Software for the FP51 programmer from ELV for the following\n");
        printf("microcontrollers: AT89C1051, AT89C2051, AT89C4051, AT89C51 and AT89C52.\n");
        printf("flip51 is free software and is distributed under the GNU Public License.\n");
        printf("flip51 comes with ABSOLUTELY NO WARRANTY; for details see the file README.\n");
        printf("Syntax: flip51 [options]\n");
        printf("Options:\n");
        printf("  -e            Erase flash ROM\n");
        printf("  -p file       Program the flash with binary file\n");
        printf("  -ph file      Program the flash with intel hex file\n");
        printf("  -r file       Read flash and store in binary file\n");
        printf("  -v file       Verify flash against binary file\n");   
        printf("  -vh file      Verify flash against intel hex file\n");   
        printf("  -l [1][2][3]  Program the specified lockbits\n");         
        printf("  -i            Show info about the uC\n");
        printf("  -io ioport    Set the address of the parallel port (in hex)\n");
        printf("                Default address is 378\n");
#ifdef TARGET_WINDOWS
        printf("  -install      Install giveio.sys driver\n");
        printf("  -remove       Remove giveio.sys driver\n");
#endif
        flip51_exit(1);
    }

#ifdef TARGET_LINUX
    
    /* get user id */
    olduid = getuid();
	
    /* setuid root for port access */
	if (setuid(0)) {perror("setuid root"); flip51_exit(1);}
	
    /* Get access to the ports */
	if (ioperm(baseport, 3, 1)) {perror("ioperm"); flip51_exit(1);}
	
    /* set original uid */
	if (setuid(0)) {perror("setuid"); flip51_exit(1);}
#endif


#ifdef TARGET_WINDOWS

	/* get windows version and check if we need the giveio driver */
	windows_version.dwOSVersionInfoSize = sizeof (OSVERSIONINFO);
	rc = GetVersionEx(&windows_version);
	if (rc == 0) {
		error = GetLastError();
        printf("error: GetVersionEx(). rc = %d LastError = %d\n", rc, error);
        flip51_exit(1);
    }

	if (windows_version.dwPlatformId == VER_PLATFORM_WIN32_NT)	/* we are on NT or better */
		use_giveio = 1;

	if (remove_giveio)
		flip51_exit(0);

	/* handle the giveio service */
	if (use_giveio) {

		/* get full path of giveio.sys */
		rc = SearchPath(NULL, "giveio.sys", NULL, sizeof giveio_path, giveio_path, &giveio_filename);
		if (rc == 0) {
			error = GetLastError();
            printf("error: cannot find the file 'giveio.sys'. LastError = %d\n", error);
            flip51_exit(1);
        }

		/* install giveio service */
        rc = PortDriverInstall(giveio_path);
        if (rc == 0) {														/* service installed sucessfully */
			giveio_installed = 1;
		} else if ((rc == -1) && (error = ERROR_ACCESS_DENIED)) {
			/* Access denied. User has no rights to install drivers. Maybe the driver is already installed? */
			/* do nothing and try to start the service */
		} else if ((rc == -2) && (error == ERROR_SERVICE_EXISTS)) {		/* service is already installed */	
			/* do nothing */
        } else {
            printf("error: cannot install giveio.sys. rc = %i  LastError = %d\n", rc, error);
            flip51_exit(1);
		}

        /* start giveio service */
		rc = PortDriverStart();
        if (rc == 0) {
			giveio_started = 1;
		} else if ((rc == -1) && (error = ERROR_ACCESS_DENIED)) {
			/* Access denied. User has no rights to start drivers. Maybe the driver is already started? */
			/* do nothing and try to open the service */
		} else if ((rc == -3) && (error == ERROR_SERVICE_ALREADY_RUNNING)) {		/* service is already running */	
			/* do nothing */
        } else {
            printf("error: cannot start giveio.sys. rc = %i  LastError = %d\n", rc, error);
            flip51_exit(1);
        }
		
		/* open giveio service */
		giveio_handle = CreateFile("\\\\.\\giveio", GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
        if (giveio_handle == INVALID_HANDLE_VALUE) {
            error = GetLastError();
			printf("error: cannot open giveio.sys. LastError = %d\n", error);
			if (error == ERROR_FILE_NOT_FOUND)
				printf("giveio.sys not installed. Use -install option to install it (you need administrator privileges to do so).\n");

            flip51_exit(1);
        }
		CloseHandle(giveio_handle);		/* we don't need it anymore */
    }              
            

    if (do_testspeed) {                  /* test computer speed */
        test_speed();
        return 0;
    }
#endif  
    

    /* check if fp51 is present */
    if (!fp51_check()) {
        printf("error: FP51 hardware not found\n");
        flip51_exit(1);
    }              

    /* reset the fp51 device */ 
    fp51_reset();

    /* read the signature byte to determine what chip is inserted */
    read_sig(&chipinfo);

    printf("Signature Bytes: %X %X %X\n", chipinfo.sig[0], chipinfo.sig[1], chipinfo.sig[2]);
    printf("Chiptype = %s, Vpp = %i Volt, flash size = %i bytes \n", chipinfo.chipname, chipinfo.volt, chipinfo.size);
    
    if (do_read) {                  /* read the chip */
        printf("Reading flash and writing to %s\n", outfilename);
        if (outfile = fopen(outfilename, "wb")) {       /* open for writing */
            read_flash(buffer, &chipinfo, chipinfo.size);
            fwrite(buffer, 1, chipinfo.size, outfile);
            fclose(outfile);
        } else {
            printf("error: cannot open %s\n", outfilename);
            flip51_exit(1);
        }
    }
    if (do_erase == 1) {                /* erase the chip */
        printf("Erasing flash\n");
        chip_erase(&chipinfo);
    }
    if (do_program) {                   /* program the chip */
        printf("Programing %s into flash\n", infilename);
        if (infile = fopen(infilename, "rb")) {
            
            filesize = read_bin_file(buffer, chipinfo.size, infile);

            if (filesize == -1)
                printf("error: file too big\n");
            else {
                printf("filesize: %i\n", filesize);
                if (write_flash(buffer, &chipinfo, filesize))
                    printf("error: programing not successful\n");
            }

            fclose(infile);
        } else {
            printf("error: cannot open %s\n", infilename);
            flip51_exit(1);
        }
    }

    if (do_program_hex) {               /* program the chip */
        printf("Programing %s into flash\n", infilename);
        if (infile = fopen(infilename, "r")) {
            
            memset(buffer, 0xff, sizeof buffer);
            filesize = read_hex_file(buffer, chipinfo.size, infile);

            if (filesize == -1)
                printf("error: file too big\n");
            else {
                printf("filesize: %i\n", filesize);
                if (write_flash(buffer, &chipinfo, filesize))
                    printf("error: programing not successful\n");
            }

            fclose(infile);
        } else {
            printf("error: cannot open %s\n", infilename);
            flip51_exit(1);
        }
    }

    if (do_verify) {                /* verify the chip */
        printf("Verifying with %s\n", vfilename);
        if (infile = fopen(vfilename, "rb")) {      /* open for reading */
            filesize = read_bin_file(vbuffer, chipinfo.size, infile);

            if (filesize == -1) {      /* file too big */
                printf("error: file too big\n");
            } else {
                read_flash(buffer, &chipinfo, filesize);
                if (memcmp (buffer, vbuffer, filesize)) {
                    printf("error: flash and file not identical\n");
                } else
                    printf("flash and file are identical\n");
            }           
            fclose(infile);
        } else {
            printf("error: cannot open %s\n", infilename);
            flip51_exit(1);
        }
    }   

    if (do_verify_hex) {                /* verify the chip */
        printf("Verifying with %s\n", vfilename);
        if (infile = fopen(vfilename, "r")) {      /* open for reading */
            memset(vbuffer, 0xff, sizeof vbuffer);
            filesize = read_hex_file(vbuffer, chipinfo.size, infile);

            if (filesize == -1) {      /* file too big */
                printf("error: file too big\n");
            } else {
                read_flash(buffer, &chipinfo, filesize);         
                if (memcmp (buffer, vbuffer, filesize)) {
                    printf("error: flash and file not identical\n");
                } else
                    printf("flash and file are identical\n");
            }           
            fclose(infile);
        } else {
            printf("error: cannot open %s\n", infilename);
            flip51_exit(1);
        }
    }   

    if (do_lockbit) {
        printf("programming lockbits ");
        if (lockbit[0])
            printf("1 ");
        if (lockbit[1])
            printf("2 ");
        if (lockbit[2])
            printf("3");
        printf("\n");
        prog_lockbits(&chipinfo);        
    }


    /* reset the fp51 device */ 
    fp51_reset();

    flip51_exit(0);
	return 0;
}

void flip51_exit(int error)
{
    int rc;
    
#ifdef TARGET_LINUX
        
    /* We don't need the ports anymore */
    if (ioperm(baseport, 3, 0)) {perror("ioperm"); flip51_exit(1);}
#endif

#ifdef TARGET_WINDOWS
	
	if (use_giveio) {

		if (install_giveio)		/* don't remove driver */
			exit(error);
		
		if (giveio_started || remove_giveio) {
			/* stop giveio service */
			rc = PortDriverStop();
			if (rc != 0) {
				printf("error: cannot stop giveio.sys. rc = %i  LastError = %d\n", rc, error);
			}
		}

		if (giveio_installed || remove_giveio) {
			/* remove giveio service */
			rc = PortDriverRemove();
			if (rc != 0) {
				printf("error: cannot remove giveio.sys. rc = %i  LastError = %i\n", rc, error);
				exit(1);
			}
		}

	}

#endif
	
    exit(error);
}



/* end of flip51.c */                                       
